"""Collections of DataArrays."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Optional, Union

import numpy as np
import pydantic.v1 as pd
import xarray as xr

from ...constants import PICOSECOND_PER_NANOMETER_PER_KILOMETER
from ...exceptions import DataError
from ...log import log
from ..base import Tidy3dBaseModel
from ..types import Axis
from .data_array import (
    DataArray,
    EMEScalarFieldDataArray,
    EMEScalarModeFieldDataArray,
    GroupIndexDataArray,
    ModeDispersionDataArray,
    ModeIndexDataArray,
    ScalarFieldDataArray,
    ScalarFieldTimeDataArray,
    ScalarModeFieldCylindricalDataArray,
    ScalarModeFieldDataArray,
    TimeDataArray,
    TriangleMeshDataArray,
)

DEFAULT_MAX_SAMPLES_PER_STEP = 10_000
DEFAULT_MAX_CELLS_PER_STEP = 10_000
DEFAULT_TOLERANCE_CELL_FINDING = 1e-6


class Dataset(Tidy3dBaseModel, ABC):
    """Abstract base class for objects that store collections of `:class:`.DataArray`s."""


class AbstractFieldDataset(Dataset, ABC):
    """Collection of scalar fields with some symmetry properties."""

    @property
    @abstractmethod
    def field_components(self) -> Dict[str, DataArray]:
        """Maps the field components to their associated data."""

    def apply_phase(self, phase: float) -> AbstractFieldDataset:
        """Create a copy where all elements are phase-shifted by a value (in radians)."""
        if phase == 0.0:
            return self
        phasor = np.exp(1j * phase)
        field_components_shifted = {}
        for fld_name, fld_cmp in self.field_components.items():
            fld_cmp_shifted = phasor * fld_cmp
            field_components_shifted[fld_name] = fld_cmp_shifted
        return self.updated_copy(**field_components_shifted)

    @property
    @abstractmethod
    def grid_locations(self) -> Dict[str, str]:
        """Maps field components to the string key of their grid locations on the yee lattice."""

    @property
    @abstractmethod
    def symmetry_eigenvalues(self) -> Dict[str, Callable[[Axis], float]]:
        """Maps field components to their (positive) symmetry eigenvalues."""

    def package_colocate_results(self, centered_fields: Dict[str, ScalarFieldDataArray]) -> Any:
        """How to package the dictionary of fields computed via self.colocate()."""
        return xr.Dataset(centered_fields)

    def colocate(self, x=None, y=None, z=None) -> xr.Dataset:
        """Colocate all of the data at a set of x, y, z coordinates.

        Parameters
        ----------
        x : Optional[array-like] = None
            x coordinates of locations.
            If not supplied, does not try to colocate on this dimension.
        y : Optional[array-like] = None
            y coordinates of locations.
            If not supplied, does not try to colocate on this dimension.
        z : Optional[array-like] = None
            z coordinates of locations.
            If not supplied, does not try to colocate on this dimension.

        Returns
        -------
        xr.Dataset
            Dataset containing all fields at the same spatial locations.
            For more details refer to `xarray's Documentation <https://tinyurl.com/cyca3krz>`_.

        Note
        ----
        For many operations (such as flux calculations and plotting),
        it is important that the fields are colocated at the same spatial locations.
        Be sure to apply this method to your field data in those cases.
        """

        if hasattr(self, "monitor") and self.monitor.colocate:
            with log as consolidated_logger:
                consolidated_logger.warning(
                    "Colocating data that has already been colocated during the solver "
                    "run. For most accurate results when colocating to custom coordinates set "
                    "'Monitor.colocate' to 'False' to use the raw data on the Yee grid "
                    "and avoid double interpolation. Note: the default value was changed to 'True' "
                    "in Tidy3D version 2.4.0."
                )

        # convert supplied coordinates to array and assign string mapping to them
        supplied_coord_map = {k: np.array(v) for k, v in zip("xyz", (x, y, z)) if v is not None}

        # dict of data arrays to combine in dataset and return
        centered_fields = {}

        # loop through field components
        for field_name, field_data in self.field_components.items():
            # loop through x, y, z dimensions and raise an error if only one element along dim
            for coord_name, coords_supplied in supplied_coord_map.items():
                coord_data = np.array(field_data.coords[coord_name])
                if coord_data.size == 1:
                    raise DataError(
                        f"colocate given {coord_name}={coords_supplied}, but "
                        f"data only has one coordinate at {coord_name}={coord_data[0]}. "
                        "Therefore, can't colocate along this dimension. "
                        f"supply {coord_name}=None to skip it."
                    )

            centered_fields[field_name] = field_data.interp(
                **supplied_coord_map, kwargs={"bounds_error": True}
            )

        # combine all centered fields in a dataset
        return self.package_colocate_results(centered_fields)


EMScalarFieldType = Union[
    ScalarFieldDataArray,
    ScalarFieldTimeDataArray,
    ScalarModeFieldDataArray,
    ScalarModeFieldCylindricalDataArray,
    EMEScalarModeFieldDataArray,
    EMEScalarFieldDataArray,
]


class ElectromagneticFieldDataset(AbstractFieldDataset, ABC):
    """Stores a collection of E and H fields with x, y, z components."""

    Ex: Optional[EMScalarFieldType] = pd.Field(
        None,
        title="Ex",
        description="Spatial distribution of the x-component of the electric field.",
    )
    Ey: Optional[EMScalarFieldType] = pd.Field(
        None,
        title="Ey",
        description="Spatial distribution of the y-component of the electric field.",
    )
    Ez: Optional[EMScalarFieldType] = pd.Field(
        None,
        title="Ez",
        description="Spatial distribution of the z-component of the electric field.",
    )
    Hx: Optional[EMScalarFieldType] = pd.Field(
        None,
        title="Hx",
        description="Spatial distribution of the x-component of the magnetic field.",
    )
    Hy: Optional[EMScalarFieldType] = pd.Field(
        None,
        title="Hy",
        description="Spatial distribution of the y-component of the magnetic field.",
    )
    Hz: Optional[EMScalarFieldType] = pd.Field(
        None,
        title="Hz",
        description="Spatial distribution of the z-component of the magnetic field.",
    )

    @property
    def field_components(self) -> Dict[str, DataArray]:
        """Maps the field components to their associated data."""
        fields = {
            "Ex": self.Ex,
            "Ey": self.Ey,
            "Ez": self.Ez,
            "Hx": self.Hx,
            "Hy": self.Hy,
            "Hz": self.Hz,
        }
        return {field_name: field for field_name, field in fields.items() if field is not None}

    @property
    def grid_locations(self) -> Dict[str, str]:
        """Maps field components to the string key of their grid locations on the yee lattice."""
        return dict(Ex="Ex", Ey="Ey", Ez="Ez", Hx="Hx", Hy="Hy", Hz="Hz")

    @property
    def symmetry_eigenvalues(self) -> Dict[str, Callable[[Axis], float]]:
        """Maps field components to their (positive) symmetry eigenvalues."""

        return dict(
            Ex=lambda dim: -1 if (dim == 0) else +1,
            Ey=lambda dim: -1 if (dim == 1) else +1,
            Ez=lambda dim: -1 if (dim == 2) else +1,
            Hx=lambda dim: +1 if (dim == 0) else -1,
            Hy=lambda dim: +1 if (dim == 1) else -1,
            Hz=lambda dim: +1 if (dim == 2) else -1,
        )


class FieldDataset(ElectromagneticFieldDataset):
    """Dataset storing a collection of the scalar components of E and H fields in the freq. domain

    Example
    -------
    >>> x = [-1,1]
    >>> y = [-2,0,2]
    >>> z = [-3,-1,1,3]
    >>> f = [2e14, 3e14]
    >>> coords = dict(x=x, y=y, z=z, f=f)
    >>> scalar_field = ScalarFieldDataArray((1+1j) * np.random.random((2,3,4,2)), coords=coords)
    >>> data = FieldDataset(Ex=scalar_field, Hz=scalar_field)
    """

    Ex: Optional[ScalarFieldDataArray] = pd.Field(
        None,
        title="Ex",
        description="Spatial distribution of the x-component of the electric field.",
    )
    Ey: Optional[ScalarFieldDataArray] = pd.Field(
        None,
        title="Ey",
        description="Spatial distribution of the y-component of the electric field.",
    )
    Ez: Optional[ScalarFieldDataArray] = pd.Field(
        None,
        title="Ez",
        description="Spatial distribution of the z-component of the electric field.",
    )
    Hx: Optional[ScalarFieldDataArray] = pd.Field(
        None,
        title="Hx",
        description="Spatial distribution of the x-component of the magnetic field.",
    )
    Hy: Optional[ScalarFieldDataArray] = pd.Field(
        None,
        title="Hy",
        description="Spatial distribution of the y-component of the magnetic field.",
    )
    Hz: Optional[ScalarFieldDataArray] = pd.Field(
        None,
        title="Hz",
        description="Spatial distribution of the z-component of the magnetic field.",
    )


class FieldTimeDataset(ElectromagneticFieldDataset):
    """Dataset storing a collection of the scalar components of E and H fields in the time domain

    Example
    -------
    >>> x = [-1,1]
    >>> y = [-2,0,2]
    >>> z = [-3,-1,1,3]
    >>> t = [0, 1e-12, 2e-12]
    >>> coords = dict(x=x, y=y, z=z, t=t)
    >>> scalar_field = ScalarFieldTimeDataArray(np.random.random((2,3,4,3)), coords=coords)
    >>> data = FieldTimeDataset(Ex=scalar_field, Hz=scalar_field)
    """

    Ex: Optional[ScalarFieldTimeDataArray] = pd.Field(
        None,
        title="Ex",
        description="Spatial distribution of the x-component of the electric field.",
    )
    Ey: Optional[ScalarFieldTimeDataArray] = pd.Field(
        None,
        title="Ey",
        description="Spatial distribution of the y-component of the electric field.",
    )
    Ez: Optional[ScalarFieldTimeDataArray] = pd.Field(
        None,
        title="Ez",
        description="Spatial distribution of the z-component of the electric field.",
    )
    Hx: Optional[ScalarFieldTimeDataArray] = pd.Field(
        None,
        title="Hx",
        description="Spatial distribution of the x-component of the magnetic field.",
    )
    Hy: Optional[ScalarFieldTimeDataArray] = pd.Field(
        None,
        title="Hy",
        description="Spatial distribution of the y-component of the magnetic field.",
    )
    Hz: Optional[ScalarFieldTimeDataArray] = pd.Field(
        None,
        title="Hz",
        description="Spatial distribution of the z-component of the magnetic field.",
    )

    def apply_phase(self, phase: float) -> AbstractFieldDataset:
        """Create a copy where all elements are phase-shifted by a value (in radians)."""

        if phase != 0.0:
            raise ValueError("Can't apply phase to time-domain field data, which is real-valued.")

        return self


class AuxFieldDataset(AbstractFieldDataset, ABC):
    """Stores a collection of aux fields with x, y, z components."""

    Nfx: Optional[EMScalarFieldType] = pd.Field(
        None,
        title="Nfx",
        description="Spatial distribution of the free carrier density for "
        "polarization in the x-direction.",
    )
    Nfy: Optional[EMScalarFieldType] = pd.Field(
        None,
        title="Nfy",
        description="Spatial distribution of the free carrier density for "
        "polarization in the y-direction.",
    )
    Nfz: Optional[EMScalarFieldType] = pd.Field(
        None,
        title="Nfz",
        description="Spatial distribution of the free carrier density for "
        "polarization in the z-direction.",
    )

    @property
    def field_components(self) -> Dict[str, DataArray]:
        """Maps the field components to their associated data."""
        fields = {
            "Nfx": self.Nfx,
            "Nfy": self.Nfy,
            "Nfz": self.Nfz,
        }
        return {field_name: field for field_name, field in fields.items() if field is not None}

    @property
    def grid_locations(self) -> Dict[str, str]:
        """Maps field components to the string key of their grid locations on the yee lattice."""
        return dict(Nfx="Ex", Nfy="Ey", Nfz="Ez")

    @property
    def symmetry_eigenvalues(self) -> Dict[str, Callable[[Axis], float]]:
        """Maps field components to their (positive) symmetry eigenvalues."""

        return dict(
            Nfx=lambda dim: +1,
            Nfy=lambda dim: +1,
            Nfz=lambda dim: +1,
        )


class AuxFieldTimeDataset(AuxFieldDataset):
    """Dataset storing a collection of the scalar components of aux fields in the time domain

    Example
    -------
    >>> x = [-1,1]
    >>> y = [-2,0,2]
    >>> z = [-3,-1,1,3]
    >>> t = [0, 1e-12, 2e-12]
    >>> coords = dict(x=x, y=y, z=z, t=t)
    >>> scalar_field = ScalarFieldTimeDataArray(np.random.random((2,3,4,3)), coords=coords)
    >>> data = AuxFieldTimeDataset(Nfx=scalar_field)
    """

    Nfx: Optional[ScalarFieldTimeDataArray] = pd.Field(
        None,
        title="Nfx",
        description="Spatial distribution of the free carrier density for polarization "
        "in the x-direction.",
    )
    Nfy: Optional[ScalarFieldTimeDataArray] = pd.Field(
        None,
        title="Nfy",
        description="Spatial distribution of the free carrier density for polarization "
        "in the y-direction.",
    )
    Nfz: Optional[ScalarFieldTimeDataArray] = pd.Field(
        None,
        title="Nfz",
        description="Spatial distribution of the free carrier density for polarization "
        "in the z-direction.",
    )


class ModeSolverDataset(ElectromagneticFieldDataset):
    """Dataset storing scalar components of E and H fields as a function of freq. and mode_index.

    Example
    -------
    >>> from tidy3d import ModeSpec
    >>> x = [-1,1]
    >>> y = [0]
    >>> z = [-3,-1,1,3]
    >>> f = [2e14, 3e14]
    >>> mode_index = np.arange(5)
    >>> field_coords = dict(x=x, y=y, z=z, f=f, mode_index=mode_index)
    >>> field = ScalarModeFieldDataArray((1+1j)*np.random.random((2,1,4,2,5)), coords=field_coords)
    >>> index_coords = dict(f=f, mode_index=mode_index)
    >>> index_data = ModeIndexDataArray((1+1j) * np.random.random((2,5)), coords=index_coords)
    >>> data = ModeSolverDataset(
    ...     Ex=field,
    ...     Ey=field,
    ...     Ez=field,
    ...     Hx=field,
    ...     Hy=field,
    ...     Hz=field,
    ...     n_complex=index_data
    ... )
    """

    Ex: Optional[ScalarModeFieldDataArray] = pd.Field(
        None,
        title="Ex",
        description="Spatial distribution of the x-component of the electric field of the mode.",
    )
    Ey: Optional[ScalarModeFieldDataArray] = pd.Field(
        None,
        title="Ey",
        description="Spatial distribution of the y-component of the electric field of the mode.",
    )
    Ez: Optional[ScalarModeFieldDataArray] = pd.Field(
        None,
        title="Ez",
        description="Spatial distribution of the z-component of the electric field of the mode.",
    )
    Hx: Optional[ScalarModeFieldDataArray] = pd.Field(
        None,
        title="Hx",
        description="Spatial distribution of the x-component of the magnetic field of the mode.",
    )
    Hy: Optional[ScalarModeFieldDataArray] = pd.Field(
        None,
        title="Hy",
        description="Spatial distribution of the y-component of the magnetic field of the mode.",
    )
    Hz: Optional[ScalarModeFieldDataArray] = pd.Field(
        None,
        title="Hz",
        description="Spatial distribution of the z-component of the magnetic field of the mode.",
    )

    n_complex: ModeIndexDataArray = pd.Field(
        ...,
        title="Propagation Index",
        description="Complex-valued effective propagation constants associated with the mode.",
    )

    n_group_raw: Optional[GroupIndexDataArray] = pd.Field(
        None,
        alias="n_group",  # This is for backwards compatibility only when loading old data
        title="Group Index",
        description="Index associated with group velocity of the mode.",
    )

    dispersion_raw: Optional[ModeDispersionDataArray] = pd.Field(
        None,
        title="Dispersion",
        description="Dispersion parameter for the mode.",
        units=PICOSECOND_PER_NANOMETER_PER_KILOMETER,
    )

    @property
    def field_components(self) -> Dict[str, DataArray]:
        """Maps the field components to their associated data."""
        fields = {
            "Ex": self.Ex,
            "Ey": self.Ey,
            "Ez": self.Ez,
            "Hx": self.Hx,
            "Hy": self.Hy,
            "Hz": self.Hz,
        }
        return {field_name: field for field_name, field in fields.items() if field is not None}

    @property
    def n_eff(self) -> ModeIndexDataArray:
        """Real part of the propagation index."""
        return self.n_complex.real

    @property
    def k_eff(self) -> ModeIndexDataArray:
        """Imaginary part of the propagation index."""
        return self.n_complex.imag

    @property
    def n_group(self) -> GroupIndexDataArray:
        """Group index."""
        if self.n_group_raw is None:
            log.warning(
                "The group index was not computed. To calculate group index, pass "
                "'group_index_step = True' in the 'ModeSpec'.",
                log_once=True,
            )
        return self.n_group_raw

    @property
    def dispersion(self) -> ModeDispersionDataArray:
        r"""Dispersion parameter.

        .. math::

           D = -\frac{\lambda}{c_0} \frac{{\rm d}^2 n_{\text{eff}}}{{\rm d}\lambda^2}
        """
        if self.dispersion_raw is None:
            log.warning(
                "The dispersion was not computed. To calculate dispersion, pass "
                "'group_index_step = True' in the 'ModeSpec'.",
                log_once=True,
            )
        return self.dispersion_raw

    def plot_field(self, *args, **kwargs):
        """Warn user to use the :class:`.ModeSolver` ``plot_field`` function now."""
        raise DeprecationWarning(
            "The 'plot_field()' method was moved to the 'ModeSolver' object."
            "Once the 'ModeSolver' is constructed, one may call '.plot_field()' on the object and "
            "the modes will be computed and displayed with 'Simulation' overlay."
        )


class PermittivityDataset(AbstractFieldDataset):
    """Dataset storing the diagonal components of the permittivity tensor.

    Example
    -------
    >>> x = [-1,1]
    >>> y = [-2,0,2]
    >>> z = [-3,-1,1,3]
    >>> f = [2e14, 3e14]
    >>> coords = dict(x=x, y=y, z=z, f=f)
    >>> sclr_fld = ScalarFieldDataArray((1+1j) * np.random.random((2,3,4,2)), coords=coords)
    >>> data = PermittivityDataset(eps_xx=sclr_fld, eps_yy=sclr_fld, eps_zz=sclr_fld)
    """

    @property
    def field_components(self) -> Dict[str, ScalarFieldDataArray]:
        """Maps the field components to their associated data."""
        return dict(eps_xx=self.eps_xx, eps_yy=self.eps_yy, eps_zz=self.eps_zz)

    @property
    def grid_locations(self) -> Dict[str, str]:
        """Maps field components to the string key of their grid locations on the yee lattice."""
        return dict(eps_xx="Ex", eps_yy="Ey", eps_zz="Ez")

    @property
    def symmetry_eigenvalues(self) -> Dict[str, Callable[[Axis], float]]:
        """Maps field components to their (positive) symmetry eigenvalues."""
        return dict(eps_xx=None, eps_yy=None, eps_zz=None)

    eps_xx: ScalarFieldDataArray = pd.Field(
        ...,
        title="Epsilon xx",
        description="Spatial distribution of the xx-component of the relative permittivity.",
    )
    eps_yy: ScalarFieldDataArray = pd.Field(
        ...,
        title="Epsilon yy",
        description="Spatial distribution of the yy-component of the relative permittivity.",
    )
    eps_zz: ScalarFieldDataArray = pd.Field(
        ...,
        title="Epsilon zz",
        description="Spatial distribution of the zz-component of the relative permittivity.",
    )


class TriangleMeshDataset(Dataset):
    """Dataset for storing triangular surface data."""

    surface_mesh: TriangleMeshDataArray = pd.Field(
        ...,
        title="Surface mesh data",
        description="Dataset containing the surface triangles and corresponding face indices "
        "for a surface mesh.",
    )


class TimeDataset(Dataset):
    """Dataset for storing a function of time."""

    values: TimeDataArray = pd.Field(
        ..., title="Values", description="Values as a function of time."
    )



"""Tool for finding and characterizing lobes in antenna radiation patterns."""

from math import isclose, isnan
from typing import Optional

import numpy as np
import pydantic.v1 as pd
from pandas import DataFrame
from scipy.signal import find_peaks, peak_widths

from ...components.base import Tidy3dBaseModel, cached_property, skip_if_fields_missing
from ...components.types import ArrayFloat1D, ArrayLike, Ax
from ...constants import fp_eps
from ...exceptions import ValidationError
from .viz import plot_params_lobe_FNBW, plot_params_lobe_peak, plot_params_lobe_width

# The minimum plateau size for peak finding, which is set to 0 to ensure that all peaks are found.
# A value must be provided to retrieve additional information from `find_peaks`.
MIN_PLATEAU_SIZE = 0
DEFAULT_MIN_LOBE_REL_HEIGHT = 1e-3
DEFAULT_NULL_THRESHOLD = 1e-3


class LobeMeasurer(Tidy3dBaseModel):
    """
    Tool for detecting and analyzing lobes in antenna radiation patterns,
    along with their characteristics such as direction and beamwidth.

    Example
    -------
    >>> theta = np.linspace(0, 2 * np.pi, 301, endpoint=False)
    >>> Urad = np.cos(theta) ** 2 * np.cos(3 * theta) ** 2
    >>> lobe_measurer = LobeMeasurer(
    ...     angle=theta,
    ...     radiation_pattern=Urad)
    >>> lobe_measures = lobe_measurer.lobe_measures
    """

    angle: ArrayFloat1D = pd.Field(
        ...,
        title="Angle",
        description="A 1-dimensional array of angles in radians. The angles should be "
        "in the range [0, 2π] and should be sorted in ascending order.",
    )

    radiation_pattern: ArrayFloat1D = pd.Field(
        ...,
        title="Radiation Pattern",
        description="A 1-dimensional array of real values representing the radiation pattern "
        "of the antenna measured on a linear scale.",
    )

    apply_cyclic_extension: bool = pd.Field(
        True,
        title="Apply Cyclic Extension",
        description="To enable accurate peak finding near boundaries of the ``angle`` array, "
        "we need to extend the signal using its periodicity. If lobes near the boundaries are not "
        "of interest, this can be set to ``False``.",
    )

    width_measure: float = pd.Field(
        0.5,
        gt=0.0,
        le=1.0,
        title="Beamwidth Measure",
        description="Relative magnitude of the lobes at which the beamwidth is measured. "
        "Default value of ``0.5`` corresponds with the half-power beamwidth.",
    )

    min_lobe_height: float = pd.Field(
        DEFAULT_MIN_LOBE_REL_HEIGHT,
        gt=0.0,
        le=1.0,
        title="Minimum Lobe Height",
        description="Only lobes in the radiation pattern with heights above this value are found. "
        "Lobe heights are measured relative to the maximum value in ``radiation_pattern``.",
    )

    null_threshold: float = pd.Field(
        DEFAULT_NULL_THRESHOLD,
        gt=0.0,
        le=1.0,
        title="Null Threshold",
        description="The threshold for detecting nulls, "
        "which is relative to the maximum value in the ``radiation_pattern``.",
    )

    @pd.validator("angle", always=True)
    def _sorted_angle(cls, val):
        """Ensure the angle array is sorted."""
        if not np.all(np.diff(val) >= 0):
            raise ValidationError("The angle array must be sorted in ascending order.")
        return val

    @pd.validator("radiation_pattern", always=True)
    def _nonnegative_radiation_pattern(cls, val):
        """Ensure the radiation pattern is nonnegative."""
        if not np.all(val >= 0):
            raise ValidationError("Radiation pattern must be nonnegative.")
        return val

    @pd.validator("apply_cyclic_extension", always=True)
    @skip_if_fields_missing(["angle"])
    def _cyclic_extension_valid(cls, val, values):
        if val:
            angle = values.get("angle")
            if np.any(angle < 0) or np.any(angle > 2 * np.pi):
                raise ValidationError(
                    "When using cyclic extension, the angle array must be in the range [0, 2π]."
                )
        return val

    @cached_property
    def lobe_measures(self) -> DataFrame:
        """
        The lobe measures as a pandas ``pandas.DataFrame`` with the following columns:

        - ``direction``: The angular position of the lobe peak.
        - ``magnitude``: The height of the lobe peak.
        - ``beamwidth``: The width of the lobe at the specified beamwidth measure.
        - ``beamwidth magnitude``: The magnitude at which the beamwidth is measured.
        - ``beamwidth bounds``: The bounds of the lobe at the the beamwidth measure.
        - ``FNBW``: The first null beam width (FNBW) of the lobe.
        - ``FNBW bounds``: The locations of the nulls on either side of the lobe.

        Returns
        -------
        DataFrame
            A DataFrame containing all lobe measures, where rows indicate the lobe index.
        """
        if self.apply_cyclic_extension:
            angle, signal = self.cyclic_extension(self.angle, self.radiation_pattern)
        else:
            angle, signal = self.angle, self.radiation_pattern

        peaks, peak_props = find_peaks(
            signal,
            plateau_size=MIN_PLATEAU_SIZE,
            height=self.min_lobe_height,
        )
        if len(peaks) == 0:
            return DataFrame()
        # Find the locations of nulls in the signal
        max_mag = np.max(signal)
        null_height = self.null_threshold * max_mag
        _, null_peak_props = find_peaks(
            -signal,
            plateau_size=MIN_PLATEAU_SIZE,
            height=-null_height,
        )

        # Extract and post process the peak properties from Scipy
        peak_locations = self._extract_peak_locations(angle, peak_props)
        peak_heights = self._extract_peak_heights(peak_props)
        widths, beamwidth_height, beamwidth_min, beamwidth_max = self._calc_peak_widths(
            angle, signal, peaks
        )
        null_locations = self._extract_peak_locations(angle, null_peak_props)
        FNBWs, FNBW_min, FNBW_max = self._first_null_beam_widths(peak_locations, null_locations)

        width_bounds = list(zip(beamwidth_min, beamwidth_max))
        FNBW_bounds = list(zip(FNBW_min, FNBW_max))
        data = {
            "direction": peak_locations,
            "magnitude": peak_heights,
            "beamwidth": widths,
            "beamwidth magnitude": beamwidth_height,
            "beamwidth bounds": width_bounds,
            "FNBW": FNBWs,
            "FNBW bounds": FNBW_bounds,
        }

        dataframe = DataFrame(data)
        if self.apply_cyclic_extension:
            dataframe = self._filter_out_of_bounds_lobes(dataframe)
        return dataframe

    @staticmethod
    def cyclic_extension(angle: ArrayLike, signal: ArrayLike) -> tuple[ArrayLike, ArrayLike]:
        """
        This helper function extends the given `signal` array by leveraging its periodic nature.
        It ensures that the peak finding algorithm in Scipy can reliably detect peaks near the
        minimum and maximum values of the ``angle`` array. The returned arrays are extended to
        the range [-π, 3π).

        Parameters
        ----------
        angle : ArrayLike
            The array of angles, expected to be in the range [0, 2π].
        signal : ArrayLike
            The array of signal values corresponding to the angle array.

        Returns
        -------
        tuple[ArrayLike, ArrayLike]
            A tuple containing the extended `angle` and `signal` arrays.
        """
        # Ensure the last sample is not duplicated if it is at 2π
        if isclose(angle[-1], 2 * np.pi, rel_tol=fp_eps):
            angle = angle[:-1]
            signal = signal[:-1]

        first_locs = np.array(np.logical_and(angle >= 0, angle < np.pi)).nonzero()
        second_locs = np.array(np.logical_and(angle >= np.pi, angle < 2 * np.pi)).nonzero()
        angle = np.concatenate(
            [angle[second_locs] - 2 * np.pi, angle, angle[first_locs] + 2 * np.pi]
        )
        signal = np.concatenate([signal[second_locs], signal, signal[first_locs]])
        return angle, signal

    @staticmethod
    def _extract_peak_locations(angle: ArrayLike, peak_props: dict) -> ArrayLike:
        """Get the peak positions in terms of the angular coordinates."""
        left_locs = peak_props["left_edges"]
        right_locs = peak_props["right_edges"]
        left_x = angle[left_locs]
        right_x = angle[right_locs]
        peaks = (left_x + right_x) / 2
        return peaks

    @staticmethod
    def _extract_peak_heights(peak_props: dict) -> ArrayLike:
        """Get the peak heights."""
        return peak_props["peak_heights"]

    def _calc_peak_widths(
        self, angle: ArrayLike, signal: ArrayLike, peaks: ArrayLike
    ) -> tuple[ArrayLike, ArrayLike, ArrayLike, ArrayLike]:
        """Get the peak widths in terms of the angular coordinates."""
        rel_height = 1.0 - self.width_measure
        last_element = len(signal) - 1
        left_ips = np.zeros_like(peaks)
        right_ips = last_element * np.ones_like(peaks)
        dummy_prominence_data = (signal[peaks], left_ips, right_ips)
        widths, width_heights, left_ips, right_ips = peak_widths(
            signal, peaks, rel_height, dummy_prominence_data
        )
        left_x = np.interp(left_ips, np.arange(len(angle)), angle)
        right_x = np.interp(right_ips, np.arange(len(angle)), angle)
        widths = right_x - left_x
        # Check that widths were properly found and are not simply the bounds
        improper_width_mask = np.logical_or(left_ips == 0, right_ips == last_element)
        widths[improper_width_mask] = np.nan
        width_heights[improper_width_mask] = np.nan
        left_x[improper_width_mask] = np.nan
        right_x[improper_width_mask] = np.nan
        return widths, width_heights, left_x, right_x

    @staticmethod
    def _first_null_beam_widths(
        peaks: ArrayLike, nulls: ArrayLike
    ) -> tuple[ArrayLike, ArrayLike, ArrayLike]:
        """Calculate the first null beam width for each lobe. If a beam null is not found,
        the value for the width is set to ``nan``."""
        insertion_indices = np.searchsorted(nulls, peaks)

        left_nulls = np.full_like(peaks, np.nan)
        right_nulls = np.full_like(peaks, np.nan)
        left_mask = (insertion_indices > 0) & (insertion_indices < len(nulls))
        right_mask = (insertion_indices >= 0) & (insertion_indices < len(nulls))
        left_nulls[left_mask] = nulls[insertion_indices[left_mask] - 1]
        right_nulls[right_mask] = nulls[insertion_indices[right_mask]]

        peak_FNBW = right_nulls - left_nulls
        return peak_FNBW, left_nulls, right_nulls

    @staticmethod
    def _filter_out_of_bounds_lobes(dataframe: DataFrame) -> DataFrame:
        """Filter out lobes that are out of bounds of the original signal."""
        filtered_dataframe = dataframe[
            dataframe["direction"].between(0, 2 * np.pi, inclusive="left")
        ]
        return filtered_dataframe.reset_index(drop=True)

    @property
    def main_lobe(self) -> DataFrame:
        """Properties of the main lobe."""
        if self.lobe_measures.empty:
            return DataFrame()
        main_lobe_idx = self.lobe_measures["magnitude"].idxmax()
        return self.lobe_measures.iloc[main_lobe_idx]

    @property
    def side_lobes(self) -> DataFrame:
        """Properties of all side lobes."""
        if self.lobe_measures.empty:
            return DataFrame()
        main_lobe_idx = self.lobe_measures["magnitude"].idxmax()
        side_lobes = self.lobe_measures.drop(main_lobe_idx)
        return side_lobes

    @property
    def sidelobe_level(self) -> Optional[float]:
        """The sidelobe level returned on a linear scale."""
        if self.side_lobes.empty:
            return None
        main_lobe_level = self.main_lobe["magnitude"]
        side_lobes = self.side_lobes
        max_side_lobe_idx = side_lobes["magnitude"].idxmax()
        side_lobe_level = side_lobes.at[max_side_lobe_idx, "magnitude"]
        return side_lobe_level / main_lobe_level

    def plot(
        self,
        lobe_index: int,
        ax: Ax,
        include_beamwidth: bool = True,
        include_FNWB: bool = True,
    ) -> Ax:
        """Annotate an existing ``matplotlib`` axis with lobe locations and widths.

        Parameters
        ----------
        lobe_index : int
            Index of the lobe to plot from the :attr:`lobe_measures`.
        ax : Ax
            ``matplotlib`` axis on which to plot the lobe measure.
        include_beamwidth : bool, optional
            If True, plot the beamwidth markers and fill the beamwidth region.
        include_FNWB : bool, optional
            If True, plot the First Null Beam Width markers and fill the FNBW region.

        Returns
        -------
        Ax
            The matplotlib axis with the plotted lobe measures.
        """

        lobe = self.lobe_measures.iloc[lobe_index]
        direction = lobe["direction"]
        ax.axvline(direction, **plot_params_lobe_peak.to_kwargs())

        # Additional annotations of the plot
        if include_beamwidth and not isnan(lobe["beamwidth magnitude"]):
            width_bounds = lobe["beamwidth bounds"]
            ax.axvline(width_bounds[0], **plot_params_lobe_width.to_kwargs())
            ax.axvline(width_bounds[1], **plot_params_lobe_width.to_kwargs())

        if include_FNWB and not isnan(lobe["FNBW"]):
            FNBW_bounds = lobe["FNBW bounds"]
            ax.axvline(FNBW_bounds[0], **plot_params_lobe_FNBW.to_kwargs())
            ax.axvline(FNBW_bounds[1], **plot_params_lobe_FNBW.to_kwargs())

        return ax



"""
This file contains a set of functions relating to packaging tidy3d for distribution. Sections of the codebase should depend on this file, but this file should not depend on any other part of the codebase.

This section should only depend on the standard core installation in the pyproject.toml, and should not depend on any other part of the codebase optional imports.
"""

import functools
from importlib import import_module
from typing import Literal

import numpy as np

from .exceptions import Tidy3dImportError

vtk = {
    "mod": None,
    "id_type": np.int64,
    "vtk_to_numpy": None,
    "numpy_to_vtkIdTypeArray": None,
    "numpy_to_vtk": None,
}


def check_import(module_name: str) -> bool:
    """
    Check if a module or submodule section has been imported. This is a functional way of loading packages that will still load the corresponding module into the total space.

    Parameters
    ----------
    module_name

    Returns
    -------
    bool
        True if the module has been imported, False otherwise.

    """
    try:
        import_module(module_name)
        return True
    except ImportError:
        return False


def verify_packages_import(modules: list, required: Literal["any", "all"] = "all"):
    def decorator(func):
        """
        When decorating a method, requires that the specified modules are available. It will raise an error if the
        module is not available depending on the value of the 'required' parameter which represents the type of
        import required.

        There are a few options to choose for the 'required' parameter:
        - 'all': All the modules must be available for the operation to continue without raising an error
        - 'any': At least one of the modules must be available for the operation to continue without raising an error

        Parameters
        ----------
        func
            The function to decorate.

        Returns
        -------
        checks_modules_import
            The decorated function.

        """

        @functools.wraps(func)
        def checks_modules_import(*args, **kwargs):
            """
            Checks if the modules are available. If they are not available, it will raise an error depending on the value.
            """
            available_modules_status = []
            maximum_amount_modules = len(modules)

            module_id_i = 0
            for module in modules:
                # Starts counting from one so that it can be compared to len(modules)
                module_id_i += 1
                import_available = check_import(module)
                available_modules_status.append(
                    import_available
                )  # Stores the status of the module import

                if not import_available:
                    if required == "all":
                        raise Tidy3dImportError(
                            f"The package '{module}' is required for this operation, but it was not found. "
                            f"Please install the '{module}' dependencies using, for example, "
                            f"'pip install tidy3d[<see_options_in_pyproject.toml>]"
                        )
                    elif required == "any":
                        # Means we need to verify that at least one of the modules is available
                        if (
                            not any(available_modules_status)
                        ) and module_id_i == maximum_amount_modules:
                            # Means that we have reached the last module and none of them were available
                            raise Tidy3dImportError(
                                f"The package '{module}' is required for this operation, but it was not found. "
                                f"Please install the '{module}' dependencies using, for example, "
                                f"'pip install tidy3d[<see_options_in_pyproject.toml>]"
                            )
                    else:
                        raise ValueError(
                            f"The value '{required}' is not a valid value for the 'required' parameter. "
                            f"Please use any 'all' or 'any'."
                        )
                else:
                    # Means that the module is available, so we can just continue with the operation
                    pass
            return func(*args, **kwargs)

        return checks_modules_import

    return decorator


def requires_vtk(fn):
    """When decorating a method, requires that vtk is available."""

    @functools.wraps(fn)
    def _fn(*args, **kwargs):
        if vtk["mod"] is None:
            try:
                import vtk as vtk_mod
                from vtk.util.numpy_support import (
                    numpy_to_vtk,
                    numpy_to_vtkIdTypeArray,
                    vtk_to_numpy,
                )
                from vtkmodules.vtkCommonCore import vtkLogger

                vtk["mod"] = vtk_mod
                vtk["vtk_to_numpy"] = vtk_to_numpy
                vtk["numpy_to_vtkIdTypeArray"] = numpy_to_vtkIdTypeArray
                vtk["numpy_to_vtk"] = numpy_to_vtk

                vtkLogger.SetStderrVerbosity(vtkLogger.VERBOSITY_WARNING)

                if vtk["mod"].vtkIdTypeArray().GetDataTypeSize() == 4:
                    vtk["id_type"] = np.int32

            except ImportError:
                raise Tidy3dImportError(
                    "The package 'vtk' is required for this operation, but it was not found. "
                    "Please install the 'vtk' dependencies using, for example, "
                    "'pip install .[vtk]'."
                )

        return fn(*args, **kwargs)

    return _fn


def get_numpy_major_version(module=np):
    """
    Extracts the major version of the installed numpy accordingly.

    Parameters
    ----------
    module : module
        The module to extract the version from. Default is numpy.

    Returns
    -------
    int
        The major version of the module.
    """
    # Get the version of the module
    module_version = module.__version__

    # Extract the major version number
    major_version = int(module_version.split(".")[0])

    return major_version



from typing import TYPE_CHECKING, Annotated, Union

from pydantic.v1 import Field

from tidy3d.components.types import TYPE_TAG_STR, ArrayLike, Complex

if TYPE_CHECKING:
    from .functions import Cos, Exp, Log, Log10, Sin, Sqrt, Tan
    from .metrics import ModeAmp, ModePower
    from .operators import (
        Abs,
        Add,
        Divide,
        FloorDivide,
        MatMul,
        Modulus,
        Multiply,
        Negate,
        Power,
        Subtract,
    )
    from .variables import Constant, Variable

NumberType = Union[int, float, Complex, ArrayLike]

OperatorType = Annotated[
    Union[
        "Add",
        "Subtract",
        "Multiply",
        "Divide",
        "Power",
        "Modulus",
        "FloorDivide",
        "MatMul",
        "Negate",
        "Abs",
    ],
    Field(discriminator=TYPE_TAG_STR),
]

FunctionType = Annotated[
    Union[
        "Sin",
        "Cos",
        "Tan",
        "Exp",
        "Log",
        "Log10",
        "Sqrt",
    ],
    Field(discriminator=TYPE_TAG_STR),
]

MetricType = Annotated[
    Union[
        "Constant",
        "Variable",
        "ModeAmp",
        "ModePower",
    ],
    Field(discriminator=TYPE_TAG_STR),
]

ExpressionType = Union[
    OperatorType,
    FunctionType,
    MetricType,
]

NumberOrExpression = Union[NumberType, ExpressionType]



"""Defines jax-compatible geometries and their conversion to grad monitors."""

from __future__ import annotations

from abc import ABC
from typing import Dict, List, Tuple, Union

import jax
import jax.numpy as jnp
import numpy as np
import pydantic.v1 as pd
import shapely
import xarray as xr
from jax.tree_util import register_pytree_node_class
from joblib import Parallel, delayed

from ....components.base import cached_property
from ....components.data.data_array import ScalarFieldDataArray
from ....components.data.monitor_data import FieldData, PermittivityData
from ....components.geometry.base import Box, Geometry, GeometryGroup
from ....components.geometry.polyslab import (
    _COMPLEX_POLYSLAB_DIVISIONS_WARN,
    _IS_CLOSE_RTOL,
    PolySlab,
)
from ....components.monitor import FieldMonitor, PermittivityMonitor
from ....components.types import ArrayFloat2D, Bound, Coordinate2D  # , annotate_type
from ....constants import MICROMETER, fp_eps
from ....exceptions import AdjointError
from ....log import log
from ...polyslab import ComplexPolySlab
from .base import WEB_ADJOINT_MESSAGE, JaxObject
from .types import JaxFloat

# number of integration points per unit wavelength in material
PTS_PER_WVL_INTEGRATION = 50

# how much to expand the gradient monitors on each side beyond the self.bounds
GRAD_MONITOR_EXPANSION = fp_eps

# maximum number of vertices allowed in JaxPolySlab
MAX_NUM_VERTICES = 1000


class JaxGeometry(Geometry, ABC):
    """Abstract :class:`.Geometry` with methods useful for all Jax subclasses."""

    @property
    def bound_size(self) -> Tuple[float, float, float]:
        """Size of the bounding box of this geometry."""
        rmin, rmax = self.bounds
        return tuple(abs(pt_max - pt_min) for (pt_min, pt_max) in zip(rmin, rmax))

    @property
    def bound_center(self) -> Tuple[float, float, float]:
        """Size of the bounding box of this geometry."""
        rmin, rmax = self.bounds

        def get_center(pt_min: float, pt_max: float) -> float:
            """Get center of bounds, including infinity, calling Geometry._get_center()."""
            pt_min = jax.lax.stop_gradient(pt_min)
            pt_max = jax.lax.stop_gradient(pt_max)
            return self._get_center(pt_min, pt_max)

        return tuple(get_center(pt_min, pt_max) for (pt_min, pt_max) in zip(rmin, rmax))

    @cached_property
    def bounding_box(self):
        """Returns :class:`JaxBox` representation of the bounding box of a :class:`JaxGeometry`.

        Returns
        -------
        :class:`JaxBox`
            Geometric object representing bounding box.
        """
        return JaxBox.from_bounds(*self.bounds)

    def make_grad_monitors(
        self, freqs: List[float], name: str
    ) -> Tuple[FieldMonitor, PermittivityMonitor]:
        """Return gradient monitor associated with this object."""
        size_enlarged = tuple(s + 2 * GRAD_MONITOR_EXPANSION for s in self.bound_size)
        field_mnt = FieldMonitor(
            size=size_enlarged,
            center=self.bound_center,
            fields=["Ex", "Ey", "Ez"],
            freqs=freqs,
            name=name + "_field",
            colocate=False,
        )

        eps_mnt = PermittivityMonitor(
            size=size_enlarged,
            center=self.bound_center,
            freqs=freqs,
            name=name + "_eps",
        )
        return field_mnt, eps_mnt

    @staticmethod
    def compute_dotted_e_d_fields(
        grad_data_fwd: FieldData, grad_data_adj: FieldData, grad_data_eps: PermittivityData
    ) -> Tuple[Dict[str, ScalarFieldDataArray], Dict[str, ScalarFieldDataArray]]:
        """Get the (x,y,z) components of E_fwd * E_adj and D_fwd * D_adj fields in the domain."""

        e_mult_xyz = {}
        d_mult_xyz = {}

        for dim in "xyz":
            # grab the E field components
            e_fld_key = f"E{dim}"
            e_fwd = grad_data_fwd.field_components[e_fld_key]
            e_adj = grad_data_adj.field_components[e_fld_key]

            # grab the epsilon data
            eps_fld_key = f"eps_{dim}{dim}"
            eps = grad_data_eps.field_components[eps_fld_key]

            # compute d fields
            d_fwd = eps * e_fwd
            d_adj = eps * e_adj

            # multiply the fwd and adj fields
            e_mult_xyz[dim] = e_fwd * e_adj
            d_mult_xyz[dim] = d_fwd * d_adj

        return e_mult_xyz, d_mult_xyz


@register_pytree_node_class
class JaxBox(JaxGeometry, Box, JaxObject):
    """A :class:`.Box` registered with jax."""

    _tidy3d_class = Box

    center_jax: Tuple[JaxFloat, JaxFloat, JaxFloat] = pd.Field(
        (0.0, 0.0, 0.0),
        title="Center (Jax)",
        description="Jax traced value for the center of the box in (x, y, z).",
        units=MICROMETER,
        stores_jax_for="center",
    )

    size_jax: Tuple[JaxFloat, JaxFloat, JaxFloat] = pd.Field(
        ...,
        title="Size (Jax)",
        description="Jax-traced value for the size of the box in (x, y, z).",
        units=MICROMETER,
        stores_jax_for="size",
    )

    def store_vjp(
        self,
        grad_data_fwd: FieldData,
        grad_data_adj: FieldData,
        grad_data_eps: PermittivityData,
        sim_bounds: Bound,
        wvl_mat: float,
        eps_out: complex,
        eps_in: complex,
        num_proc: int = 1,
    ) -> JaxBox:
        """Stores the gradient of the box parameters given forward and adjoint field data."""

        # multiply the E and D field components together for fwd and adj
        e_mult_xyz, d_mult_xyz = self.compute_dotted_e_d_fields(
            grad_data_fwd=grad_data_fwd, grad_data_adj=grad_data_adj, grad_data_eps=grad_data_eps
        )

        rmin, rmax = bounds_intersect = self.bounds_intersection(self.bounds, sim_bounds)

        # stores vjps for the min and max surfaces on all dimensions
        vjp_surfs = {dim: np.array([0.0, 0.0]) for dim in "xyz"}

        # loop through all 6 surfaces (x,y,z) & (-, +)
        for dim_index, dim_normal in enumerate("xyz"):
            for min_max_index, min_max_val in enumerate(bounds_intersect):
                # get the normal coordinate of this surface
                normal_coord = {dim_normal: min_max_val[dim_index]}

                # skip if the geometry edge is out of bounds of the simulation
                sim_min_max_val = sim_bounds[min_max_index][dim_index]
                geo_min_max_val = self.bounds[min_max_index][dim_index]
                if (min_max_index == 0) and (geo_min_max_val <= sim_min_max_val):
                    continue
                if (min_max_index == 1) and (geo_min_max_val >= sim_min_max_val):
                    continue

                # get the dimensions and edge values on the plane of this surface
                _, dims_plane = self.pop_axis("xyz", axis=dim_index)
                _, mins_plane = self.pop_axis(rmin, axis=dim_index)
                _, maxs_plane = self.pop_axis(rmax, axis=dim_index)

                # construct differential area value and coordinates evenly spaced along this surface
                d_area = 1.0
                area_coords = {}
                for dim_plane, min_edge, max_edge in zip(dims_plane, mins_plane, maxs_plane):
                    # if there is no thickness along this dimension, skip it
                    length_edge = max_edge - min_edge
                    if length_edge == 0:
                        continue

                    num_cells_dim = int(length_edge * PTS_PER_WVL_INTEGRATION / wvl_mat) + 1

                    # update the differential area value
                    d_len = length_edge / num_cells_dim
                    d_area *= d_len

                    # construct evenly spaced coordinates along this dimension
                    interp_vals = np.linspace(
                        min_edge + d_len / 2, max_edge - d_len / 2, num_cells_dim
                    )
                    area_coords[dim_plane] = interp_vals

                # for each field component
                for field_cmp_dim in "xyz":
                    # select the permittivity data
                    eps_field_name = f"eps_{field_cmp_dim}{field_cmp_dim}"
                    eps_data = grad_data_eps.field_components[eps_field_name]

                    # get the permittivity values just inside and outside the edge

                    num_cells_normal_dim = len(eps_data.coords[dim_normal])

                    # if the eps_data is <=3 grid cells thick, pick the middle cell
                    if num_cells_normal_dim <= 3:
                        isel_ins = num_cells_normal_dim // 2

                    # otherwise, pick the cell 4 pixels from outside cell
                    else:
                        n_cells_in = 3
                        isel_ins = n_cells_in if min_max_index == 0 else -n_cells_in - 1

                    isel_out = 0 if min_max_index == 0 else -1

                    eps2 = eps_data.isel(**{dim_normal: isel_out})
                    eps1 = eps_data.isel(**{dim_normal: isel_ins})

                    # get gradient contribution for normal component using normal D field
                    if field_cmp_dim == dim_normal:
                        # construct normal D fields, dotted together at surface
                        d_normal = d_mult_xyz[field_cmp_dim]
                        d_normal = d_normal.interp(**normal_coord, assume_sorted=True)

                        # compute adjoint contribution using perturbation theory for shifting bounds
                        delta_eps_inv = 1.0 / eps1 - 1.0 / eps2
                        d_integrand = -(delta_eps_inv * d_normal).real
                        d_integrand = d_integrand.interp(**area_coords, assume_sorted=True)
                        grad_contrib = d_area * np.sum(d_integrand.values)

                    # get gradient contribution for parallel components using parallel E fields
                    else:
                        # measure parallel E fields, dotted together at surface
                        e_parallel = e_mult_xyz[field_cmp_dim]
                        e_parallel = e_parallel.interp(**normal_coord, assume_sorted=True)

                        # compute adjoint contribution using perturbation theory for shifting bounds
                        delta_eps = eps1 - eps2
                        e_integrand = +(delta_eps * e_parallel).real
                        e_integrand = e_integrand.interp(**area_coords, assume_sorted=True)
                        grad_contrib = d_area * np.sum(e_integrand.values)

                    # add this field contribution to the dict storing the surface contributions
                    vjp_surfs[dim_normal][min_max_index] += grad_contrib

        # convert surface vjps to center, size vjps. Note, convert these to jax types w/ np.sum()
        vjp_center = tuple(np.sum(vjp_surfs[dim][1] - vjp_surfs[dim][0]) for dim in "xyz")
        vjp_size = tuple(np.sum(0.5 * (vjp_surfs[dim][1] + vjp_surfs[dim][0])) for dim in "xyz")
        return self.copy(update=dict(center_jax=vjp_center, size_jax=vjp_size))


@register_pytree_node_class
class JaxPolySlab(JaxGeometry, PolySlab, JaxObject):
    """A :class:`.PolySlab` registered with jax."""

    _tidy3d_class = PolySlab

    vertices_jax: Tuple[Tuple[JaxFloat, JaxFloat], ...] = pd.Field(
        ...,
        title="Vertices (Jax)",
        description="Jax-traced list of (d1, d2) defining the 2 dimensional positions of the "
        "polygon face vertices at the ``reference_plane``. "
        "The index of dimension should be in the ascending order: e.g. if "
        "the slab normal axis is ``axis=y``, the coordinate of the vertices will be in (x, z)",
        units=MICROMETER,
        stores_jax_for="vertices",
    )

    slab_bounds_jax: Tuple[JaxFloat, JaxFloat] = pd.Field(
        ...,
        title="Slab bounds (Jax)",
        description="Jax-traced list of (h1, h2) defining the minimum and maximum positions "
        "of the slab along the ``axis`` dimension. ",
        units=MICROMETER,
        stores_jax_for="slab_bounds",
    )

    sidewall_angle_jax: JaxFloat = pd.Field(
        default=0.0,
        title="Sidewall angle (Jax)",
        description="Jax-traced float defining the sidewall angle of the slab "
        "along the ``axis`` dimension. ",
        units=MICROMETER,
        stores_jax_for="sidewall_angle",
    )

    dilation_jax: JaxFloat = pd.Field(
        default=0.0,
        title="Dilation (Jax)",
        description="Jax-traced float defining the dilation.",
        units=MICROMETER,
        stores_jax_for="dilation",
    )

    @pd.validator("sidewall_angle", always=True)
    def no_sidewall(cls, val):
        """Warn if sidewall angle present."""
        if not np.isclose(val, 0.0):
            log.warning(
                "'JaxPolySlab' does not yet perform the full adjoint gradient treatment "
                "for slanted sidewalls. "
                "A straight sidewall angle is assumed when computing the gradient with respect "
                "to shifting boundaries of the geometry. Therefore, as 'sidewall_angle' becomes "
                "further from '0.0', the gradient error can be significant. "
                "If high gradient accuracy is needed, please either reduce your 'sidewall_angle' "
                "or wait until this feature is supported fully in a later version.",
                log_once=True,
            )
        return val

    def _validate_web_adjoint(self) -> None:
        """Run validators for this component, only if using ``tda.web.run()``."""
        self._limit_number_of_vertices()

    def _limit_number_of_vertices(self) -> None:
        """Limit the maximum number of vertices."""
        if len(self.vertices_jax) > MAX_NUM_VERTICES:
            raise AdjointError(
                f"For performance, a maximum of {MAX_NUM_VERTICES} are allowed in 'JaxPolySlab'. "
                + WEB_ADJOINT_MESSAGE
            )

    def _extrusion_length_to_offset_distance(self, extrusion: float) -> float:
        """Convert extrusion length to offset distance."""
        if jnp.isclose(self.sidewall_angle_jax, 0):
            return 0.0
        return -extrusion * self._tanq

    @staticmethod
    def _orient(vertices: jnp.ndarray) -> jnp.ndarray:
        """Return a CCW-oriented polygon.

        Parameters
        ----------
        vertices : np.ndarray
            Shape (N, 2) defining the polygon vertices in the xy-plane.

        Returns
        -------
        np.ndarray
            Vertices of a CCW-oriented polygon.
        """
        return vertices if JaxPolySlab._area(vertices) > 0 else vertices[::-1, :]

    @staticmethod
    def _area(vertices: jnp.ndarray) -> float:
        """Compute the signed polygon area (positive for CCW orientation).

        Parameters
        ----------
        vertices : np.ndarray
            Shape (N, 2) defining the polygon vertices in the xy-plane.

        Returns
        -------
        float
            Signed polygon area (positive for CCW orientation).
        """
        x = vertices[:, 0]
        y = vertices[:, 1]
        return jnp.dot(x, jnp.roll(y, -1)) - jnp.dot(y, jnp.roll(x, -1)) / 2

    @staticmethod
    def _shift_vertices(
        vertices: jnp.ndarray, dist
    ) -> Tuple[jnp.ndarray, jnp.ndarray, Tuple[jnp.ndarray, jnp.ndarray]]:
        """Shifts the vertices of a polygon outward uniformly by distances
        `dists`.

        Parameters
        ----------
        jnp.ndarray
            Shape (N, 2) defining the polygon vertices in the xy-plane.
        dist : float
            Distance to offset.

        Returns
        -------
        Tuple[jnp.ndarray, jnp.narray, Tuple[jnp.ndarray, jnp.ndarray]]
            New polygon vertices;
            and the shift of vertices in direction parallel to the edges.
            Shift along x and y direction.
        """

        if jnp.isclose(dist, 0):
            return vertices, jnp.zeros(vertices.shape[0], dtype=float), None

        def rot90(v):
            return jnp.array([-v[1], v[0]])

        def normalize(v):
            return v / jnp.linalg.norm(v, axis=0)

        vs = vertices.T
        vs_next = jnp.roll(vs, axis=-1, shift=-1)
        vs_previous = jnp.roll(vs, axis=-1, shift=+1)

        asp = normalize(vs_next - vs)
        asm = normalize(vs - vs_previous)

        # the vertex shift is decomposed into parallel and perpendicular directions
        perpendicular_shift = -dist
        det = jnp.cross(asm, asp, axis=0)

        tan_half_angle = jnp.where(
            jnp.isclose(det, 0, rtol=_IS_CLOSE_RTOL),
            0.0,
            jnp.cross(asm, rot90(asm - asp), axis=0)
            / (det + jnp.isclose(det, 0, rtol=_IS_CLOSE_RTOL)),
        )
        parallel_shift = dist * tan_half_angle

        shift_total = perpendicular_shift * rot90(asm) + parallel_shift * asm

        return jnp.swapaxes(vs + shift_total, -2, -1), parallel_shift, shift_total

    @staticmethod
    def _neighbor_vertices_crossing_detection(
        vertices: jnp.ndarray, dist: float, ignore_at_dist: bool = True
    ) -> float:
        """Detect if neighboring vertices will cross after a dilation distance dist.

        Parameters
        ----------
        vertices : jnp.ndarray
            Shape (N, 2) defining the polygon vertices in the xy-plane.
        dist : float
            Distance to offset.
        ignore_at_dist : bool, optional
            whether to ignore the event right at ``dist`.

        Returns
        -------
        float
            the absolute value of the maximal allowed dilation
            if there are any crossing, otherwise return ``None``.
        """
        # ignore the event that occurs right at the offset distance
        if ignore_at_dist:
            dist -= fp_eps * dist / jnp.abs(dist)

        edge_length, edge_reduction = JaxPolySlab._edge_length_and_reduction_rate(vertices)
        length_remaining = edge_length - edge_reduction * dist

        mask = length_remaining < 0
        if jnp.any(mask):
            return jnp.min(jnp.abs(edge_length[mask] / edge_reduction[mask]))
        return None

    @staticmethod
    def _edge_length_and_reduction_rate(vertices: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
        """Edge length of reduction rate of each edge with unit offset length.

        Parameters
        ----------
        vertices : jnp.ndarray
            Shape (N, 2) defining the polygon vertices in the xy-plane.

        Returns
        -------
        Tuple[jnp.ndarray, jnp.narray]
            edge length, and reduction rate
        """

        # edge length
        vs = vertices.T
        vs_next = jnp.roll(vs, axis=-1, shift=-1)
        edge_length = jnp.linalg.norm(vs_next - vs, axis=0)

        # edge length remaining
        parallel_shift = JaxPolySlab._shift_vertices(vertices, 1.0)[1]
        parallel_shift_p = jnp.roll(parallel_shift, shift=-1)
        edge_reduction = -(parallel_shift + parallel_shift_p)
        return edge_length, edge_reduction

    @staticmethod
    def _remove_duplicate_vertices(vertices: jnp.ndarray) -> jnp.ndarray:
        """Remove redundant/identical nearest neighbour vertices.

        Parameters
        ----------
        vertices : jnp.ndarray
            Shape (N, 2) defining the polygon vertices in the xy-plane.

        Returns
        -------
        np.ndarray
            Vertices of polygon.
        """

        vertices_f = jnp.roll(vertices, shift=-1, axis=0)
        vertices_diff = jnp.linalg.norm(vertices - vertices_f, axis=1)
        return vertices[~jnp.isclose(vertices_diff, 0, rtol=_IS_CLOSE_RTOL)]

    @staticmethod
    def _proper_vertices(vertices: ArrayFloat2D) -> jnp.ndarray:
        """convert vertices to jnp.array format,
        removing duplicate neighbouring vertices,
        and oriented in CCW direction.

        Returns
        -------
        ArrayLike[float, float]
           The vertices of the polygon for internal use.
        """

        vertices_np = JaxPolySlab.vertices_to_array(vertices)
        return JaxPolySlab._orient(JaxPolySlab._remove_duplicate_vertices(vertices_np))

    @staticmethod
    def _heal_polygon(vertices: jnp.ndarray) -> jnp.ndarray:
        """heal a self-intersecting polygon."""
        shapely_poly = PolySlab.make_shapely_polygon(jax.lax.stop_gradient(vertices))
        if shapely_poly.is_valid:
            return vertices

        raise NotImplementedError(
            "The dilation caused damage to the polygon. Automatically healing this is "
            "currently not supported for 'JaxPolySlab' objects. Try increasing the spacing "
            "between vertices or reduce the amount of dilation."
        )

    @staticmethod
    def vertices_to_array(vertices_tuple: ArrayFloat2D) -> jnp.ndarray:
        """Converts a list of tuples (vertices) to a jax array."""
        return jnp.asarray(vertices_tuple)

    @cached_property
    def reference_polygon(self) -> jnp.ndarray:
        """The polygon at the reference plane.

        Returns
        -------
        ArrayLike[float, float]
            The vertices of the polygon at the reference plane.
        """
        vertices = JaxPolySlab._proper_vertices(self.vertices_jax)
        if jnp.isclose(self.dilation, 0):
            return vertices
        offset_vertices = self._shift_vertices(vertices, self.dilation)[0]
        return self._heal_polygon(offset_vertices)

    def edge_contrib(
        self,
        vertex_grad: Coordinate2D,
        vertex_stat: Coordinate2D,
        is_next: bool,
        e_mult_xyz: Tuple[Dict[str, ScalarFieldDataArray]],
        d_mult_xyz: Tuple[Dict[str, ScalarFieldDataArray]],
        sim_bounds: Bound,
        wvl_mat: float,
        eps_out: complex,
        eps_in: complex,
    ) -> Coordinate2D:
        """Gradient w.r.t change in ``vertex_grad`` connected to ``vertex_stat``."""

        # TODO: (later) compute these grabbing from grad_data_eps at some distance away
        delta_eps_12 = eps_in - eps_out
        delta_eps_inv_12 = 1.0 / eps_in - 1.0 / eps_out

        # get edge pointing from moving vertex to static vertex
        vertex_grad = np.array(jax.lax.stop_gradient(vertex_grad))
        vertex_stat = np.array(jax.lax.stop_gradient(vertex_stat))
        edge = vertex_stat - vertex_grad
        length_edge = np.linalg.norm(edge)

        # if the edge length is 0 (overlapping vertices), there is no gradient contrib for this edge
        if np.isclose(length_edge, 0.0):
            return 0.0

        # get normalized vectors tangent to and perpendicular to edge in global caresian basis
        tx, ty = edge / length_edge
        normal_vector = np.array((+ty, -tx))

        # ensure normal vector is pointing "out" assuming clockwise vertices
        if is_next:
            normal_vector *= -1
        nx, ny = normal_vector

        # Check if vertices CCW or CW. Multiply by +1 if CCW to ensure normal out
        if self.is_ccw:
            normal_vector *= -1

        def edge_position(s: np.array) -> np.array:
            """Parameterization of position along edge from s=0 (static) to s=1 (gradient)."""
            return (1 - s) * vertex_stat[:, None] + s * vertex_grad[:, None]

        def edge_basis(
            xyz_components: Tuple[FieldData, FieldData, FieldData],
        ) -> Tuple[FieldData, FieldData, FieldData]:
            """Puts a field component from the (x, y, z) basis to the (t, n, z) basis."""
            cmp_z, (cmp_x_edge, cmp_y_edge) = self.pop_axis(xyz_components, axis=self.axis)

            cmp_t = cmp_x_edge * tx + cmp_y_edge * ty
            cmp_n = cmp_x_edge * nx + cmp_y_edge * ny

            return cmp_t, cmp_n, cmp_z

        def compute_integrand(s: np.array, z: np.array) -> np.array:
            """Get integrand at positions ``(s, z)`` along the edge."""

            # grab the position along edge and make dictionary of coords to interp with (s, z)
            x, y = edge_position(s=s)
            x = xr.DataArray(x, coords={"s": s})
            y = xr.DataArray(y, coords={"s": s})
            coords_interp = dict(x=x, y=y, z=z)

            def evaluate(scalar_field: ScalarFieldDataArray) -> float:
                """Evaluate a scalar field at a coordinate along the edge."""

                # if only 1 z coordinate, just isel the data.
                if len(z) == 1:
                    scalar_field = scalar_field.isel(z=0)
                    coords_xy = {key: coords_interp[key] for key in "xy"}
                    return scalar_field.interp(**coords_xy, assume_sorted=True)

                return scalar_field.interp(**coords_interp, assume_sorted=True)

            e_xyz_eval = [evaluate(e_fld) for e_fld in e_mult_xyz.values()]
            d_xyz_eval = [evaluate(d_fld) for d_fld in d_mult_xyz.values()]

            e_t_edge, _, e_z_edge = edge_basis(xyz_components=e_xyz_eval)
            _, d_n_edge, _ = edge_basis(xyz_components=d_xyz_eval)

            # get the correct sign to apply to the new fields
            sign_t, sign_n, sign_z = edge_basis(xyz_components=(1.0, 1.0, 1.0))
            e_t_edge *= sign_t
            d_n_edge *= sign_n
            e_z_edge *= sign_z

            # multiply by the change in epsilon (in, out) terms and sum contributions
            contrib_e_t = delta_eps_12 * e_t_edge
            contrib_e_z = delta_eps_12 * e_z_edge
            contrib_d_n = -delta_eps_inv_12 * d_n_edge
            contrib_total = contrib_e_t + contrib_d_n + contrib_e_z

            # scale the gradient contribution by the normalized distance from the static edge
            # make broadcasting work with both 2D and 3D simulation domains
            return (s * contrib_total.T).T

        # discretize along the edge
        # TODO: handle edge case where a vertex lies far outside simulation domain
        num_cells_edge = int(length_edge * PTS_PER_WVL_INTEGRATION / wvl_mat) + 1
        ds = 1.0 / float(num_cells_edge)
        s_vals = np.linspace(0 + ds / 2, 1 - ds / 2, num_cells_edge)

        # find the min and max of the slab within the simulation bounds
        slab_min, slab_max = self.slab_bounds
        sim_rmin, sim_rmax = sim_bounds
        sim_min = sim_rmin[self.axis]
        sim_max = sim_rmax[self.axis]
        z_max = min(slab_max, sim_max)
        z_min = max(slab_min, sim_min)

        # discretize along z
        length_axis = abs(z_max - z_min)
        num_cells_axis = int(length_axis * PTS_PER_WVL_INTEGRATION / wvl_mat) + 1
        dz = float(length_axis) / float(num_cells_axis)

        # handle a 2D simulation along axis (unitless)
        z_vals = np.linspace(z_min + dz / 2, z_max - dz / 2, num_cells_axis)

        if dz == 0.0:
            dz = 1.0

        # integrate by summing over axis edge (z) and parameterization point (s)
        integrand = compute_integrand(s=s_vals, z=z_vals).sum(dim="f")
        integral_result = np.sum(integrand.fillna(0).values)

        # project to the normal direction
        integral_result *= normal_vector

        # take the real part (from adjoint math) and multiply by area element
        return length_edge * ds * dz * np.real(integral_result)

    def vertex_vjp(
        self,
        i_vertex,
        e_mult_xyz: Tuple[Dict[str, ScalarFieldDataArray]],
        d_mult_xyz: Tuple[Dict[str, ScalarFieldDataArray]],
        sim_bounds: Bound,
        wvl_mat: float,
        eps_out: complex,
        eps_in: complex,
    ):
        """Compute the vjp for every vertex."""

        # get the location of the "previous" and "next" vertices in the polygon
        vertex = self.vertices[i_vertex]
        vertex_prev = self.vertices[(i_vertex - 1) % len(self.vertices)]
        vertex_next = self.vertices[(i_vertex + 1) % len(self.vertices)]

        # taking the current vertex "static", compute the edge contributions from prev and next
        contrib_next = self.edge_contrib(
            vertex_grad=vertex,
            vertex_stat=vertex_next,
            is_next=True,
            e_mult_xyz=e_mult_xyz,
            d_mult_xyz=d_mult_xyz,
            sim_bounds=sim_bounds,
            wvl_mat=wvl_mat,
            eps_out=eps_out,
            eps_in=eps_in,
        )
        contrib_prev = self.edge_contrib(
            vertex_grad=vertex,
            vertex_stat=vertex_prev,
            is_next=False,
            e_mult_xyz=e_mult_xyz,
            d_mult_xyz=d_mult_xyz,
            sim_bounds=sim_bounds,
            wvl_mat=wvl_mat,
            eps_out=eps_out,
            eps_in=eps_in,
        )

        # add the "forward" contribution from the "previous" contribution to get the vertex VJP
        return contrib_prev + contrib_next

    def store_vjp(
        self,
        grad_data_fwd: FieldData,
        grad_data_adj: FieldData,
        grad_data_eps: PermittivityData,
        sim_bounds: Bound,
        wvl_mat: float,
        eps_out: complex,
        eps_in: complex,
        num_proc: int = 1,
    ) -> JaxPolySlab:
        """Stores the gradient of the vertices given forward and adjoint field data."""

        # multiply the E and D field components together for fwd and adj
        e_mult_xyz, d_mult_xyz = self.compute_dotted_e_d_fields(
            grad_data_fwd=grad_data_fwd, grad_data_adj=grad_data_adj, grad_data_eps=grad_data_eps
        )

        if num_proc is not None and num_proc > 1:
            return self.store_vjp_parallel(
                e_mult_xyz=e_mult_xyz,
                d_mult_xyz=d_mult_xyz,
                sim_bounds=sim_bounds,
                wvl_mat=wvl_mat,
                eps_out=eps_out,
                eps_in=eps_in,
                num_proc=num_proc,
            )

        return self.store_vjp_sequential(
            e_mult_xyz=e_mult_xyz,
            d_mult_xyz=d_mult_xyz,
            sim_bounds=sim_bounds,
            wvl_mat=wvl_mat,
            eps_out=eps_out,
            eps_in=eps_in,
        )

    def _make_vertex_args(
        self,
        e_mult_xyz: FieldData,
        d_mult_xyz: FieldData,
        sim_bounds: Bound,
        wvl_mat: float,
        eps_out: complex,
        eps_in: complex,
    ) -> tuple:
        """Generate arguments for ``vertex_vjp``."""

        num_verts = len(self.vertices)

        arg_list = []

        for i in range(num_verts):
            args_i = [i] + [e_mult_xyz, d_mult_xyz, sim_bounds, wvl_mat, eps_out, eps_in]
            arg_list.append(args_i)

        return arg_list

    def store_vjp_sequential(
        self,
        e_mult_xyz: FieldData,
        d_mult_xyz: FieldData,
        sim_bounds: Bound,
        wvl_mat: float,
        eps_out: complex,
        eps_in: complex,
    ) -> JaxPolySlab:
        """Stores the gradient of the vertices given forward and adjoint field data."""
        # Construct arguments to pass to the parallel vertices_vjp computation

        arg_list = self._make_vertex_args(
            e_mult_xyz, d_mult_xyz, sim_bounds, wvl_mat, eps_out, eps_in
        )
        vertices_vjp = tuple(self.vertex_vjp(*args) for args in arg_list)
        vertices_vjp = tuple(tuple(x) for x in vertices_vjp)

        return self.updated_copy(vertices_jax=vertices_vjp)

    def store_vjp_parallel(
        self,
        e_mult_xyz: FieldData,
        d_mult_xyz: FieldData,
        sim_bounds: Bound,
        wvl_mat: float,
        eps_out: complex,
        eps_in: complex,
        num_proc: int = 1,
    ) -> JaxPolySlab:
        """Stores the gradient of the vertices given forward and adjoint field data."""

        args = self._make_vertex_args(e_mult_xyz, d_mult_xyz, sim_bounds, wvl_mat, eps_out, eps_in)
        vertices_vjp = Parallel(n_jobs=num_proc)(delayed(self.vertex_vjp)(*arg) for arg in args)
        vertices_vjp = tuple(tuple(x) for x in vertices_vjp)
        return self.updated_copy(vertices_jax=vertices_vjp)


@register_pytree_node_class
class JaxComplexPolySlab(JaxPolySlab, ComplexPolySlab):
    """A :class:`.ComplexPolySlab` registered with jax."""

    _tidy3d_class = ComplexPolySlab

    @pd.validator("vertices", always=True)
    def no_self_intersecting_polygon_during_extrusion(cls, val, values):
        """Turn off the validation for this class."""
        return val

    @property
    def geometry_group(self) -> None:
        """Divide a complex jax polyslab into a list of simple polyslabs, which
        are assembled into a :class:`.JaxGeometryGroup`.

        Returns
        -------
        :class:`.JaxGeometryGroup`
            JaxGeometryGroup for a list of simple jax polyslabs divided from the complex
            polyslab.
        """
        return JaxGeometryGroup(geometries=self.sub_polyslabs)

    def _dilation_value_at_reference_to_coord(self, dilation: float) -> float:
        """Compute the coordinate based on the dilation value to the reference plane."""

        z_coord = -dilation / self._tanq + self.slab_bounds_jax[0]
        if self.reference_plane == "middle":
            return z_coord + self.finite_length_axis / 2
        if self.reference_plane == "top":
            return z_coord + self.finite_length_axis
        # bottom case
        return z_coord

    @property
    def sub_polyslabs(self) -> List[JaxPolySlab]:
        """Divide a complex polyslab into a list of simple polyslabs.
        Only neighboring vertex-vertex crossing events are treated in this
        version.

        Returns
        -------
        List[JaxPolySlab]
            A list of simple jax polyslabs.
        """
        sub_polyslab_list = []
        num_division_count = 0

        # initialize sub-polyslab parameters

        sub_polyslab_dict = self.dict(
            exclude={  # all of these NEED to be overwritten, so best to exclude them
                "type",
                "vertices",
                "vertices_jax",
                "slab_bounds",
                "slab_bounds_jax",
                "sidewall_angle",
                "sidewall_angle_jax",
                "dilation",
                "reference_plane",
            }
        )
        if jnp.isclose(self.sidewall_angle_jax, 0):
            return [
                JaxPolySlab(
                    vertices=tuple(map(tuple, self.vertices_jax)),
                    slab_bounds=tuple(self.slab_bounds_jax),
                    sidewall_angle=self.sidewall_angle_jax,
                    dilation=self.dilation,
                    reference_plane=self.reference_plane,
                    **sub_polyslab_dict,
                )
            ]

        # initialize offset distance
        offset_distance = 0.0

        for dist_val in self._dilation_length:
            dist_now = 0.0
            vertices_now = self.reference_polygon

            # constructing sub-polyslabs until reaching the base/top
            while not jnp.isclose(dist_now, dist_val):
                # bounds for sub-polyslabs assuming no self-intersection
                slab_bounds = [
                    self._dilation_value_at_reference_to_coord(dist_now),
                    self._dilation_value_at_reference_to_coord(dist_val),
                ]
                # 1) find out any vertices touching events between the current
                # position to the base/top
                max_dist = JaxPolySlab._neighbor_vertices_crossing_detection(
                    vertices_now, dist_val - dist_now
                )

                # vertices touching events captured, update bounds for sub-polyslab
                if max_dist is not None:
                    # max_dist doesn't have sign, so construct signed offset distance
                    offset_distance = max_dist * dist_val / jnp.abs(dist_val)
                    slab_bounds = [
                        self._dilation_value_at_reference_to_coord(dist_now),
                        self._dilation_value_at_reference_to_coord(dist_now + offset_distance),
                    ]

                # 2) construct sub-polyslab
                slab_bounds = jnp.sort(
                    jnp.asarray(slab_bounds)
                )  # for reference_plane=top/bottom, bounds need to be ordered
                # direction of marching
                reference_plane = "bottom" if dist_val / self._tanq < 0 else "top"

                sub_polyslab_list.append(
                    JaxPolySlab(
                        vertices=tuple(map(tuple, vertices_now)),
                        slab_bounds=tuple(slab_bounds),
                        sidewall_angle=self.sidewall_angle_jax,
                        dilation=0.0,  # dilation accounted for in setup
                        reference_plane=reference_plane,
                        **sub_polyslab_dict,
                    )
                )

                # Now Step 3
                if max_dist is None:
                    break
                dist_now += offset_distance
                # new polygon vertices where collapsing vertices are removed but keep one
                vertices_now = JaxPolySlab._shift_vertices(vertices_now, offset_distance)[0]
                vertices_now = JaxPolySlab._remove_duplicate_vertices(vertices_now)
                # all vertices collapse
                if len(vertices_now) < 3:
                    break
                # polygon collapse into 1D
                if shapely.Polygon(jax.lax.stop_gradient(vertices_now)).buffer(0).area < fp_eps:
                    raise RuntimeError("Unhandled shapely transformation in JaxComplexPolySlab.")
                vertices_now = JaxPolySlab._orient(vertices_now)
                num_division_count += 1

        if num_division_count > _COMPLEX_POLYSLAB_DIVISIONS_WARN:
            log.warning(
                f"Too many self-intersecting events: the polyslab has been divided into "
                f"{num_division_count} polyslabs; more than {_COMPLEX_POLYSLAB_DIVISIONS_WARN} may "
                f"slow down the simulation."
            )

        return sub_polyslab_list


JaxSingleGeometryType = Union[JaxBox, JaxPolySlab]


@register_pytree_node_class
class JaxGeometryGroup(JaxGeometry, GeometryGroup, JaxObject):
    """A collection of Geometry objects that can be called as a single geometry object."""

    _tidy3d_class = GeometryGroup

    geometries: Tuple[JaxPolySlab, ...] = pd.Field(
        ...,
        title="Geometries",
        description="Tuple of jax geometries in a single grouping. "
        "Can provide significant performance enhancement in ``JaxStructure`` when all geometries "
        "are assigned the same ``JaxMedium``. Note: at this moment, only ``JaxPolySlab`` "
        "is supported.",
        jax_field=True,
    )

    @staticmethod
    def _store_vjp_geometry(
        geometry: JaxSingleGeometryType,
        grad_data_fwd: FieldData,
        grad_data_adj: FieldData,
        grad_data_eps: PermittivityData,
        sim_bounds: Bound,
        wvl_mat: float,
        eps_out: complex,
        eps_in: complex,
    ) -> JaxSingleGeometryType:
        """Function to store a single vjp for a single geometry."""
        return geometry.store_vjp(
            grad_data_fwd=grad_data_fwd,
            grad_data_adj=grad_data_adj,
            grad_data_eps=grad_data_eps,
            sim_bounds=sim_bounds,
            wvl_mat=wvl_mat,
            eps_out=eps_out,
            eps_in=eps_in,
        )

    def store_vjp(
        self,
        grad_data_fwd: FieldData,
        grad_data_adj: FieldData,
        grad_data_eps: PermittivityData,
        sim_bounds: Bound,
        wvl_mat: float,
        eps_out: complex,
        eps_in: complex,
        num_proc: int = 1,
    ) -> JaxGeometryGroup:
        """Returns a ``JaxGeometryGroup`` where the ``.geometries`` store the gradient info."""

        args_list = []
        for geo in self.geometries:
            args_i = [
                geo,
                grad_data_fwd,
                grad_data_adj,
                grad_data_eps,
                sim_bounds,
                wvl_mat,
                eps_out,
                eps_in,
            ]
            args_list.append(args_i)

        if num_proc == 1:
            geometries_vjp = tuple(self._store_vjp_geometry(*args) for args in args_list)
        else:
            geometries_vjp = tuple(
                Parallel(n_jobs=num_proc)(
                    delayed(self._store_vjp_geometry)(*args) for args in args_list
                )
            )

        return self.updated_copy(geometries=geometries_vjp)


JaxGeometryType = Union[JaxSingleGeometryType, JaxGeometryGroup]


JAX_GEOMETRY_MAP = {
    Box: JaxBox,
    PolySlab: JaxPolySlab,
    ComplexPolySlab: JaxComplexPolySlab,
    GeometryGroup: JaxGeometryGroup,
}



"""Tidy3d abstraction types for the core."""

from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum

from pydantic.v1 import BaseModel


class Tidy3DResource(BaseModel, ABC):
    """Abstract base class / template for a webservice that implements resource query."""

    @classmethod
    @abstractmethod
    def get(cls, *args, **kwargs) -> Tidy3DResource:
        """Get a resource from the server."""


class ResourceLifecycle(Tidy3DResource, ABC):
    """Abstract base class for a webservice that implements resource life cycle management."""

    @classmethod
    @abstractmethod
    def create(cls, *args, **kwargs) -> Tidy3DResource:
        """Create a new resource and return it."""

    @abstractmethod
    def delete(self, *args, **kwargs) -> None:
        """Delete the resource."""


class Submittable(BaseModel, ABC):
    """Abstract base class / template for a webservice that implements a submit method."""

    @abstractmethod
    def submit(self, *args, **kwargs) -> None:
        """Submit the task to the webservice."""


class Queryable(BaseModel, ABC):
    """Abstract base class / template for a webservice that implements a query method."""

    @classmethod
    @abstractmethod
    def list(cls, *args, **kwargs) -> [Queryable]:
        """List all resources of this type."""


class TaskType(str, Enum):
    FDTD = "FDTD"
    MODE_SOLVER = "MODE_SOLVER"
    HEAT = "HEAT"
    HEAT_CHARGE = "HEAT_CHARGE"
    EME = "EME"
    MODE = "MODE"



"""
Utility functions for the tidy3d develop CLI.
"""

import pathlib
import subprocess

import tidy3d

__all__ = [
    "get_install_directory",
    "echo_and_run_subprocess",
    "echo_and_check_subprocess",
]


def get_install_directory():
    """
    Retrieve the installation directory of the tidy3d module.

    Returns
    -------
    pathlib.Path
        The absolute path of the parent directory of the tidy3d module.
    """
    return pathlib.Path(tidy3d.__file__).parent.parent.absolute()


def echo_and_run_subprocess(command: list, **kwargs):
    """
    Print and execute a subprocess command.

    Parameters
    ----------
    command : list
        A list of command line arguments to be executed.
    **kwargs : dict
        Additional keyword arguments to pass to subprocess.run.

    Returns
    -------
    subprocess.CompletedProcess
        The result of the subprocess execution.
    """
    concatenated_command = " ".join(command)
    print("Running: " + concatenated_command)
    return subprocess.run(command, cwd=get_install_directory(), **kwargs)


def echo_and_check_subprocess(command: list, *args, **kwargs):
    """
    Print and execute a subprocess command, ensuring it completes successfully.

    Parameters
    ----------
    command : list
        A list of command line arguments to be executed.
    **kwargs : dict
        Additional keyword arguments to pass to subprocess.check_call.

    Returns
    -------
    int
        The return code of the subprocess execution.
    """
    concatenated_command = " ".join(command)
    print("Running: " + concatenated_command)
    return subprocess.check_call(command, *args, **kwargs, cwd=get_install_directory())



import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.constants import PERCMCUBE, VOLT


# Band-gap narrowing models
class SlotboomBandGapNarrowing(Tidy3dBaseModel):
    """
    Parameters for the Slotboom model for band-gap narrowing.

    Notes
    ------
        The Slotboom band-gap narrowing :math:`\\Delta E_G` model is discussed in [1]_ as follows:

        .. math::

            \\Delta E_G = V_{1,bgn} \\left( \\ln \\left( \\frac{N_{tot}}{N_{2,bgn}} \\right)
            + \\sqrt{\\left( \\ln \\left( \\frac{N_{tot}}{N_{2,bgn}} \\right) \\right)^2 + C_{2,bgn}} \\right)
            \\quad \\text{if} \\quad N_{tot} \\geq 10^{15} \\text{cm}^{-3},

            \\Delta E_G = 0 \\quad \\text{if} \\quad N_{tot} < 10^{15} \\text{cm}^{-3}.

        Note that :math:`N_{tot}` is the total doping as defined within a :class:`SemiconductorMedium`.

        Example
        -------
            >>> import tidy3d as td
            >>> default_Si = td.SlotboomBandGapNarrowing(
            ...    v1=6.92 * 1e-3,
            ...    n2=1.3e17,
            ...    c2=0.5,
            ...    min_N=1e15,
            ... )

        .. [1] 'UNIFIED APPARENT BANDGAP NARROWING IN n- AND p-TYPE SILICON' Solid-State Electronics Vol. 35, No. 2, pp. 125-129, 1992"""

    v1: pd.PositiveFloat = pd.Field(
        ...,
        title=r"$V_{1,bgn}$ parameter",
        description=r"$V_{1,bgn}$ parameter",
        units=VOLT,
    )

    n2: pd.PositiveFloat = pd.Field(
        ...,
        title=r"$N_{2,bgn}$ parameter",
        description=r"$N_{2,bgn}$ parameter",
        units=PERCMCUBE,
    )

    c2: float = pd.Field(
        title=r"$C_{2,bgn}$ parameter",
        description=r"$C_{2,bgn}$ parameter",
    )

    min_N: pd.NonNegativeFloat = pd.Field(
        ...,
        title="Minimum total doping",
        description="Bandgap narrowing is applied at location where total doping "
        "is higher than 'min_N'.",
        units=PERCMCUBE,
    )



"""
Note in the future we might want to implement interpolation models here.
"""

from typing import Union

from tidy3d.components.material.tcad.charge import (
    ChargeConductorMedium,
    ChargeInsulatorMedium,
    SemiconductorMedium,
)
from tidy3d.components.material.tcad.heat import ThermalSpecType
from tidy3d.components.medium import MediumType, MediumType3D

OpticalMediumType = MediumType
ElectricalMediumType = MediumType
HeatMediumType = ThermalSpecType
ChargeMediumType = Union[ChargeConductorMedium, ChargeInsulatorMedium, SemiconductorMedium]

OpticalMediumType3D = ElectricalMediumType3D = ChargeMediumType3D = MediumType3D



"""Class and custom data array for representing a scattering-matrix port, which is defined by a pair of terminals."""

from abc import ABC, abstractmethod
from typing import Union

import pydantic.v1 as pd

from ....components.base import Tidy3dBaseModel, cached_property
from ....components.data.data_array import FreqDataArray
from ....components.data.sim_data import SimulationData
from ....components.grid.grid import Grid
from ....components.monitor import FieldMonitor, ModeMonitor
from ....components.source.base import Source
from ....components.source.time import GaussianPulse
from ....components.types import FreqArray
from ....log import log


class AbstractTerminalPort(Tidy3dBaseModel, ABC):
    """Class representing a single terminal-based port. All terminal ports must provide methods
    for computing voltage and current. These quantities represent the voltage between the
    terminals, and the current flowing from one terminal into the other.
    """

    name: str = pd.Field(
        ...,
        title="Name",
        description="Unique name for the port.",
        min_length=1,
    )

    @cached_property
    @abstractmethod
    def injection_axis(self):
        """Injection axis of the port."""

    @abstractmethod
    def to_source(
        self, source_time: GaussianPulse, snap_center: float = None, grid: Grid = None
    ) -> Source:
        """Create a current source from a terminal-based port."""

    def to_field_monitors(
        self, freqs: FreqArray, snap_center: float = None, grid: Grid = None
    ) -> Union[list[FieldMonitor], list[ModeMonitor]]:
        """DEPRECATED: Monitors used to compute the port voltage and current."""
        log.warning(
            "'to_field_monitors' method name is deprecated and will be removed in the future. Please use "
            "'to_monitors' for the same effect."
        )
        return self.to_monitors(freqs=freqs, snap_center=snap_center, grid=grid)

    @abstractmethod
    def to_monitors(
        self, freqs: FreqArray, snap_center: float = None, grid: Grid = None
    ) -> Union[list[FieldMonitor], list[ModeMonitor]]:
        """Monitors used to compute the port voltage and current."""

    @abstractmethod
    def compute_voltage(self, sim_data: SimulationData) -> FreqDataArray:
        """Helper to compute voltage across the port."""

    @abstractmethod
    def compute_current(self, sim_data: SimulationData) -> FreqDataArray:
        """Helper to compute current flowing into the port."""



"""Find corners of structures on a 2D plane."""

from typing import List, Literal, Optional

import numpy as np
import pydantic.v1 as pd

from ...constants import inf
from ..base import Tidy3dBaseModel
from ..geometry.base import Box, ClipOperation
from ..geometry.utils import merging_geometries_on_plane
from ..medium import PEC, LossyMetalMedium
from ..structure import Structure
from ..types import ArrayFloat2D, Axis

CORNER_ANGLE_THRESOLD = 0.1 * np.pi


class CornerFinderSpec(Tidy3dBaseModel):
    """Specification for corner detection on a 2D plane."""

    medium: Literal["metal", "dielectric", "all"] = pd.Field(
        "metal",
        title="Material Type For Corner Identification",
        description="Find corners of structures made of ``medium``, "
        "which can take value ``metal`` for PEC and lossy metal, ``dielectric`` "
        "for non-metallic materials, and ``all`` for all materials.",
    )

    angle_threshold: float = pd.Field(
        CORNER_ANGLE_THRESOLD,
        title="Angle Threshold In Corner Identification",
        description="A vertex is qualified as a corner if the angle spanned by its two edges "
        "is larger than the supplementary angle of "
        "this threshold value.",
        ge=0,
        lt=np.pi,
    )

    distance_threshold: Optional[pd.PositiveFloat] = pd.Field(
        None,
        title="Distance Threshold In Corner Identification",
        description="If not ``None`` and the distance of the vertex to its neighboring vertices "
        "is below the threshold value based on Douglas-Peucker algorithm, the vertex is disqualified as a corner.",
    )

    def corners(
        self,
        normal_axis: Axis,
        coord: float,
        structure_list: List[Structure],
    ) -> ArrayFloat2D:
        """On a 2D plane specified by axis = `normal_axis` and coordinate `coord`, find out corners of merged
        geometries made of `medium`.


        Parameters
        ----------
        normal_axis : Axis
            Axis normal to the 2D plane.
        coord : float
            Position of plane along the normal axis.
        structure_list : List[Structure]
            List of structures present in simulation.

        Returns
        -------
        ArrayFloat2D
            Corner coordinates.
        """

        # Construct plane
        center = [0, 0, 0]
        size = [inf, inf, inf]
        center[normal_axis] = coord
        size[normal_axis] = 0
        plane = Box(center=center, size=size)

        # prepare geometry and medium list
        geometry_list = [structure.geometry for structure in structure_list]
        # For metal, we don't distinguish between LossyMetal and PEC,
        # so they'll be merged to PEC. Other materials are considered as dielectric.
        medium_list = (structure.medium for structure in structure_list)
        medium_list = [
            PEC if (mat.is_pec or isinstance(mat, LossyMetalMedium)) else mat for mat in medium_list
        ]
        # merge geometries
        merged_geos = merging_geometries_on_plane(geometry_list, plane, medium_list)

        # corner finder
        corner_list = []
        for mat, shapes in merged_geos:
            if self.medium != "all" and mat.is_pec != (self.medium == "metal"):
                continue
            polygon_list = ClipOperation.to_polygon_list(shapes)
            for poly in polygon_list:
                poly = poly.normalize().buffer(0)
                if self.distance_threshold is not None:
                    poly = poly.simplify(self.distance_threshold, preserve_topology=True)
                corner_list.append(self._filter_collinear_vertices(list(poly.exterior.coords)))
                # in case the polygon has holes
                for poly_inner in poly.interiors:
                    corner_list.append(self._filter_collinear_vertices(list(poly_inner.coords)))

        if len(corner_list) > 0:
            corner_list = np.concatenate(corner_list)
        return corner_list

    def _filter_collinear_vertices(self, vertices: ArrayFloat2D) -> ArrayFloat2D:
        """Filter collinear vertices of a polygon, and return corners.

        Parameters
        ----------
        vertices : ArrayFloat2D
            Polygon vertices from shapely.Polygon. The last vertex is identical to the 1st
            vertex to make a valid polygon.

        Returns
        -------
        ArrayFloat2D
            Corner coordinates.
        """

        def normalize(v):
            return v / np.linalg.norm(v, axis=-1)[:, np.newaxis]

        # drop the last vertex, which is identical to the 1st one.
        vs_orig = np.array(vertices[:-1])
        # compute unit vector to next and previous vertex
        vs_next = np.roll(vs_orig, axis=0, shift=-1)
        vs_previous = np.roll(vs_orig, axis=0, shift=+1)
        unit_next = normalize(vs_next - vs_orig)
        unit_previous = normalize(vs_previous - vs_orig)
        # angle
        inner_product = np.sum(unit_next * unit_previous, axis=-1)
        inner_product = np.where(inner_product > 1, 1, inner_product)
        inner_product = np.where(inner_product < -1, -1, inner_product)
        angle = np.arccos(inner_product)
        ind_filter = angle <= np.pi - self.angle_threshold
        return vs_orig[ind_filter]



"""Defines heat-charge material specifications for 'HeatChargeSimulation'"""

from __future__ import annotations

from abc import ABC
from typing import Tuple

import pydantic.v1 as pd

from tidy3d.components.base import cached_property
from tidy3d.components.base_sim.source import AbstractSource
from tidy3d.components.tcad.viz import plot_params_heat_source
from tidy3d.components.viz import PlotParams
from tidy3d.exceptions import SetupError


class AbstractHeatChargeSource(AbstractSource, ABC):
    """Abstract source for heat-charge simulations. All source types
    for 'HeatChargeSimulation' derive from this class."""

    @cached_property
    def plot_params(self) -> PlotParams:
        """Default parameters for plotting a Source object."""
        return plot_params_heat_source


class StructureBasedHeatChargeSource(AbstractHeatChargeSource):
    """Abstract class associated with structures. Sources associated
    to structures must derive from this class"""

    structures: Tuple[str, ...] = pd.Field(
        title="Target Structures",
        description="Names of structures where to apply heat source.",
    )

    @pd.validator("structures", always=True)
    def check_non_empty_structures(cls, val):
        """Error if source doesn't point at any structures."""
        if len(val) == 0:
            raise SetupError("List of structures for heat source is empty.")

        return val


class GlobalHeatChargeSource(AbstractHeatChargeSource):
    """Abstract heat/charge source applied to all structures in the simulation"""
